/*
	Hyperion - A network anomaly discovery system
	Copyright (C) 2011 Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "config.h"
#include "error.hpp"

#include "buffer.hpp"

Buffer::Buffer()
{
	data = 0;
	length = 0;
	offset = 0;
}

Buffer::Buffer(unsigned int length)
{
	data = 0; //This has to be so, because of a check performed by GetBuffer()
	offset = 0;
	GetBuffer(length);
}

Buffer::~Buffer()
{
	if (data){
		data->refcount--;
		if (!data->refcount){
			delete [](data->data);
			delete data;
		}
	}
}

Buffer::Buffer(const Buffer &rhs)
{
	data = rhs.data;
	length = rhs.length;
	offset = rhs.offset;
	if (data){
		data->refcount++;
	}
}

Buffer &Buffer::operator=(const Buffer &rhs)
{
	if (this != &rhs){
		Data *olddata = data;
		data = rhs.data;
		length = rhs.length;
		offset = rhs.offset;
		if (data){
			data->refcount++;
		}
		
		if (olddata){
			olddata->refcount--;
			if (!olddata->refcount){
				delete [](olddata->data);
				delete olddata;
			}
		}
	}
	return *this;
}

Buffer::Buffer(const unsigned char *pdata, unsigned int plength)
{
	data = new Data;
	data->data = new unsigned char[plength];
	for (unsigned int i = 0; i < plength; i++){
		data->data[i] = pdata[i];
	}
	data->refcount = 1;
	length = plength;
	offset = 0;
}

Buffer::Buffer(const Buffer &b, unsigned int o)
{
	if (b.data == 0){
		throw FormatError("Cannot copy from empty buffer");
	}
	if (b.length < o){
		throw FormatError("Attempt to increase buffer by %u; beyond length %u", o, b.length);
	}
	data = b.data;
	data->refcount++;
	offset = b.offset + o;
	length = b.length - o;
}

Buffer::Buffer(const Buffer &b, unsigned int o, unsigned int l)
{
	if (b.data == 0){
		throw FormatError("Cannot copy from empty buffer");
	}
	if (b.length < o + l){
		throw FormatError("Attempt to create buffer of length %u from buffer of length %u with offset %u",
			l, b.length, o);
	}
	data = b.data;
	data->refcount++;
	offset = b.offset + o;
	length = l;
}


unsigned char *Buffer::GetBuffer(unsigned int plength)
{
	if (data){
		throw FormatError("Cannot get buffer for non-empty buffer");
	}
	data = new Data;
	data->data = new unsigned char[plength];
	data->refcount = 1;
	length = plength;
	return data->data;
}

Buffer Buffer::operator+(const Buffer &rhs) const
{
	if (!data){
		return rhs;
	}
	if (!rhs.data){
		return *this;
	}
	
	Buffer b;
	unsigned char *d = b.GetBuffer(length + rhs.length);
	for (unsigned int i = 0; i < length; i++){
		d[i] = data->data[i + offset];
	}
	for (unsigned int i = 0; i < rhs.length; i++){
		d[length + i] = rhs.data->data[rhs.offset + i];
	}
	b.offset = 0;
	b.length = length + rhs.length;
	return b;
}

//This is useful for having buffers in maps
bool Buffer::operator< (const Buffer &rhs) const
{
	if (!data){ //empty buffers come before everything else
		return rhs.data;
	}
	
	if (!rhs.data){ //empty buffers come before everything else
		return false;
	}
	
	if (length < rhs.length){
		return true; //shorter sequences are always before longer ones
	}
	else if (length > rhs.length){
		return false;
	}
	
	//both buffers are of the same length
	for (unsigned int i = 0; i < length; i++){
		if (data->data[i + offset] < rhs.data->data[i + rhs.offset]){
			return true;
		}
		else if (data->data[i + offset] > rhs.data->data[i + rhs.offset]){
			return false;
		}
	}
	
	//they're actually equal
	return false;
}

bool Buffer::operator==(const Buffer &rhs) const
{
	if (!data && !rhs.data){
		return true;
	}
	if (length != rhs.length){
		return false;
	}
	for (unsigned int i = 0; i < length; i++){
		if (data->data[i + offset] != rhs.data->data[i + rhs.offset]){
			return false;
		}
	}
	return true;
}

//Calculates the checksum, as specified by RFC 1071, and places it at
//the specified address. http://tools.ietf.org/html/rfc1071
void Buffer::Rfc1071(unsigned int address)
{
	BoundsCheck(address + 1);
	data->data[address] = 0;
	data->data[address + 1] = 0;
	uint16_t n = GetRfc1071();
	data->data[address] = n & 0xFF;
	data->data[address + 1] = n >> 8;
}

//True if the buffer checksums correctly
bool Buffer::Rfc1071() const
{
	return (GetRfc1071() == 0xFFFF);
}

uint16_t Buffer::GetRfc1071() const
{
	unsigned int checksum = 0;
	for (uint32_t i = 0; i < (length + 1)/2; i++){
		uint32_t n = 0;
		n = (uint32_t)data->data[offset + (i*2)];
		
		if (i * 2 != length - 1){
			n += ((uint32_t)data->data[offset + (i*2) + 1] << 8);
		}
		
		checksum += n;
	}
	
	while (checksum & 0xFFFF0000){
		checksum = (checksum & 0xFFFF) + (checksum >> 16);
	}
	
	return ~checksum;
}

unsigned int Buffer::WriteToBuffer(unsigned char *b, unsigned int l)
{
	if (!data){
		return 0;
	}
	unsigned int n = length < l ? length : l;
	for (unsigned int i = 0; i < n; i++){
		b[i] = data->data[i + offset];
	}
	return n;
}

void Buffer::CopyOnWrite()
{
	if (data->refcount > 1){
		Data *tmp = new Data;
		tmp->refcount = 1;
		tmp->data = new unsigned char[length];
		for (unsigned int i = 0; i < length; i++){
			tmp->data[i] = data->data[i + offset];
		}
		data->refcount--;
		data = tmp;
		offset = 0;
	}
}

