/*
	Hyperion - A network anomaly discovery system
	Copyright (C) 2011 Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "config.h"
#include "error.hpp"

#ifndef BUFFER_HPP
#define BUFFER_HPP

#include "config.h"
#include "error.hpp"

#include <stdint.h>
#include <arpa/inet.h>

struct Data
{
	unsigned char *data;
	unsigned int refcount;
};

class Buffer
{
	public:
		Buffer();
		Buffer(unsigned int length);
		Buffer(const Buffer &rhs);
		Buffer(const unsigned char *pdata, unsigned int plength);
		Buffer(const Buffer &b, unsigned int o);
		Buffer(const Buffer &b, unsigned int o, unsigned int l);
		Buffer &operator=(const Buffer &rhs);
		virtual ~Buffer();
		
		inline const unsigned char &operator[](unsigned int address) const
		{
			BoundsCheck(address);
			return data->data[address + offset];
		}
		
		inline unsigned int GetLength()
		{
			return length;
		}
		
		inline uint8_t  GetUInt8 (unsigned int address) const
		{
			BoundsCheck(address);
			return data->data[address + offset];
		}
		
		inline uint16_t GetUInt16(unsigned int address) const
		{
			BoundsCheck(address + 1);
			return *((uint16_t *)(data->data + address + offset));
		}
		
		inline uint32_t GetUInt32(unsigned int address) const
		{
			BoundsCheck(address + 3);
			return *((uint32_t *)(data->data + address + offset));
		}
		
		inline uint64_t GetUInt64(unsigned int address) const
		{
			BoundsCheck(address + 7);
			return *((uint64_t *)(data->data + address + offset));
		}
		
		inline uint16_t GetUInt16N(unsigned int address) const
		{
			return ntohs(GetUInt16(address));
		}
		
		inline uint32_t GetUInt32N(unsigned int address) const
		{
			return ntohl(GetUInt32(address));
		}
		
		inline void SetUInt8 (unsigned int address, uint8_t v)
		{
			BoundsCheck(address);
			CopyOnWrite();
			data->data[address + offset] = v;
		}
		
		inline void SetUInt16(unsigned int address, uint16_t v)
		{
			BoundsCheck(address + 1);
			CopyOnWrite();
			*((uint16_t *)(data->data + address + offset)) = v;
		}
		
		inline void SetUInt32(unsigned int address, uint32_t v)
		{
			BoundsCheck(address + 3);
			CopyOnWrite();
			*((uint32_t *)(data->data + address + offset)) = v;
		}
		
		inline void SetUInt64(unsigned int address, uint64_t v)
		{
			BoundsCheck(address + 7);
			CopyOnWrite();
			*((uint64_t *)(data->data + address + offset)) = v;
		}
		
		inline void SetUInt16N(unsigned int address, uint16_t v)
		{
			SetUInt16(address, htons(v));
		}
		
		inline void GetUInt32N(unsigned int address, uint32_t v)
		{
			SetUInt32(address, htonl(v));
		}
		
		inline const unsigned char *GetBuffer() const
		{
			return data->data;
		}
		
		inline Buffer &operator+=(const Buffer &rhs)
		{
			Buffer tmp = *this;
			*this = tmp + rhs;
			return *this;
		}
		
		inline bool operator!=(const Buffer &rhs) const
		{
			return !(*this == rhs);
		}
		
		unsigned char *GetBuffer(unsigned int plength);
		Buffer operator+(const Buffer &rhs) const;
		bool operator< (const Buffer &rhs) const;
		bool operator==(const Buffer &rhs) const;
		void Rfc1071(unsigned int address);
		bool Rfc1071() const;
		uint16_t GetRfc1071() const;
		unsigned int WriteToBuffer(unsigned char *b, unsigned int l);
		
		//TODO: virtual ToString() to do DynamicValue's ToString, or an overridden one
		
	private:
		inline void BoundsCheck(unsigned int address) const
		{
			#ifdef DEBUG
				if (!data){
					throw FormatError("Cannot read from empty buffer");
				}
				if (address >= length){
					throw FormatError("%u is out of bounds for buffer of length %u", address, length);
				}
			#endif
		}
		void CopyOnWrite();
		
		Data *data;
		unsigned int offset;
		unsigned int length;
};

#endif