/*
	Hyperion - A network anomaly discovery system
	Copyright (C) 2011 Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "config.h"
#include "error.hpp"

#include "configuration.hpp"

ConfigurationAstNodeExpression::~ConfigurationAstNodeExpression(){}

ConfigurationAstNodeAssignment::ConfigurationAstNodeAssignment(std::string pname, ConfigurationAstNodeExpression *pexpression)
{
	name = pname;
	expression = pexpression;
}
#include <cstdio>
ConfigurationAstNodeAssignment::~ConfigurationAstNodeAssignment()
{
	delete expression;
}

void ConfigurationAstNodeAssignment::Execute(DynamicValue &root, std::string modulename)
{
	root[modulename][name] = expression->GetValue(root, modulename);
}


ConfigurationAstNodeModule::ConfigurationAstNodeModule
	(std::string pname, std::vector<ConfigurationAstNodeAssignment *> passignments)
{
	name = pname;
	assignments = passignments;
}

ConfigurationAstNodeModule::~ConfigurationAstNodeModule()
{
	for (std::vector<ConfigurationAstNodeAssignment *>::iterator it = assignments.begin(); it != assignments.end(); it++){
		delete *it;
	}
}

void ConfigurationAstNodeModule::Execute(DynamicValue &root)
{
	for (std::vector<ConfigurationAstNodeAssignment *>::iterator it = assignments.begin(); it != assignments.end(); it++){
		(*it)->Execute(root, name);
	}
}


ConfigurationAstNodeRoot::ConfigurationAstNodeRoot(std::vector<ConfigurationAstNodeModule *> pmodules)
{
	modules = pmodules;
}

ConfigurationAstNodeRoot::~ConfigurationAstNodeRoot()
{
	for (std::vector<ConfigurationAstNodeModule *>::iterator it = modules.begin(); it != modules.end(); it++){
		delete *it;
	}
}

void ConfigurationAstNodeRoot::Execute(DynamicValue &root)
{
	for (std::vector<ConfigurationAstNodeModule *>::iterator it = modules.begin(); it != modules.end(); it++){
		(*it)->Execute(root);
	}
}

ConfigurationAstNodeOperator::ConfigurationAstNodeOperator
	(char pop, ConfigurationAstNodeExpression *pa, ConfigurationAstNodeExpression *pb)
{
	op = pop;
	a = pa;
	b = pb;
}

ConfigurationAstNodeOperator::~ConfigurationAstNodeOperator()
{
	delete a;
	delete b;
}

DynamicValue ConfigurationAstNodeOperator::GetValue(DynamicValue &root, std::string modulename)
{
	return a->GetValue(root, modulename).Op(op, b->GetValue(root, modulename));
}


ConfigurationAstNodeValue::ConfigurationAstNodeValue(DynamicValue pvalue)
{
	value = pvalue;
}

DynamicValue ConfigurationAstNodeValue::GetValue(DynamicValue &root, std::string modulename)
{
	return value;
}


ConfigurationAstNodeVariable::ConfigurationAstNodeVariable(std::string pname)
{
	name = pname;
}

DynamicValue ConfigurationAstNodeVariable::GetValue(DynamicValue &root, std::string modulename)
{
	return root[modulename][name];
}

ConfigurationAstNodeValueList::ConfigurationAstNodeValueList(std::vector<ConfigurationAstNodeExpression *> pexpressions)
{
	expressions = pexpressions;
}

ConfigurationAstNodeValueList::~ConfigurationAstNodeValueList()
{
	for (std::vector<ConfigurationAstNodeExpression *>::iterator it = expressions.begin(); it != expressions.end(); it++){
		delete *it;
	}
}

DynamicValue ConfigurationAstNodeValueList::GetValue(DynamicValue &root, std::string modulename)
{
	std::vector<DynamicValue> values;
	for (std::vector<ConfigurationAstNodeExpression *>::iterator it = expressions.begin(); it != expressions.end(); it++){
		values.push_back((*it)->GetValue(root, modulename));
	}
	return DynamicValue(values);
}
