/*
	Hyperion - A network anomaly discovery system
	Copyright (C) 2011-2012 Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "config.h"
#include "error.hpp"

#define __STDC_FORMAT_MACROS
#include <inttypes.h>

#include "dynamic.hpp"
#include "event.hpp"

static std::string GetIndent(unsigned int depth)
{
	std::string indent = "";
	for (unsigned int i = 0; i < depth; i++){
		indent += "\t";
	}
	return indent;
}

static std::string Indent(std::string s, unsigned int depth)
{
	std::string indent = GetIndent(depth);
	std::string r = "";
	for (std::string::iterator it = s.begin(); it != s.end(); it++){
		r += *it;
		if (*it == '\n'){
			r += indent;
		}
	}
	return r;
}


class DynamicValueDataBase
{
	public:
		virtual ~DynamicValueDataBase()
		{
		}
		
		virtual DynamicValueDataBase *copy() = 0;
		
		virtual DynamicValueDataBase *Op(char op, DynamicValueDataBase *rhs) = 0;
		
		
		virtual operator bool()
		{
			throw FormatError("Cannot convert %s to bool", GetValueType());
		}
		
		virtual operator int64_t()
		{
			throw FormatError("Cannot convert %s to integer", GetValueType());
		}
		
		virtual operator void *()
		{
			throw FormatError("Cannot convert %s to pointer", GetValueType());
		}
		
		virtual operator double()
		{
			throw FormatError("Cannot convert %s to double", GetValueType());
		}
		
		virtual operator std::string()
		{
			throw FormatError("Cannot convert %s to string", GetValueType());
		}
		
		virtual operator Time()
		{
			throw FormatError("Cannot convert %s to time", GetValueType());
		}
		
		virtual operator Buffer()
		{
			throw FormatError("Cannot convert %s to buffer", GetValueType());
		}
		
		virtual operator Event()
		{
			throw FormatError("Cannot convert %s to event", GetValueType());
		}
		
		virtual operator std::vector<DynamicValue> &()
		{
			throw FormatError("Cannot convert %s to list", GetValueType());
		}
		
		virtual operator std::map<std::string, DynamicValue> &()
		{
			throw FormatError("Cannot convert %s to map", GetValueType());
		}
		
		virtual std::string ToString() = 0;
		
	private:
		virtual const char *GetValueType() = 0;
};

//You'd think these could be templated, but the specialization becomes EVIL;
//Don't do it unless you find a sane way
class DynamicValueDataBool : public DynamicValueDataBase
{
	public:
		DynamicValueDataBool(bool pv)
		{
			v = pv;
		}
		
		virtual DynamicValueDataBase *copy()
		{
			return new DynamicValueDataBool(*this);
		}
		
		virtual DynamicValueDataBase *Op(char op, DynamicValueDataBase *rhs)
		{
			if (!dynamic_cast<DynamicValueDataBool *>(rhs)){
				throw FormatError("Left hand side and right hand side are of different types");
			}
			throw FormatError("Cannot perform operator %c on bools");
		}
		
		virtual operator bool()
		{
			return v;
		}
		
		virtual std::string ToString()
		{
			return v ? "true" : "false";
		}
		
	private:
		virtual const char *GetValueType()
		{
			return "bool";
		}
		
		bool v;
};

class DynamicValueDataPointer : public DynamicValueDataBase
{
	public:
		DynamicValueDataPointer(void *pv)
		{
			v = pv;
		}
		
		virtual DynamicValueDataBase *copy()
		{
			return new DynamicValueDataPointer(*this);
		}
		
		virtual DynamicValueDataBase *Op(char op, DynamicValueDataBase *rhs)
		{
			if (!dynamic_cast<DynamicValueDataBool *>(rhs)){
				throw FormatError("Left hand side and right hand side are of different types");
			}
			throw FormatError("Cannot perform operator %c on bools");
		}
		
		virtual operator void *()
		{
			return v;
		}
		
		virtual std::string ToString()
		{
			return "POINTER";
		}
		
	private:
		virtual const char *GetValueType()
		{
			return "pointer";
		}
		
		void *v;
};

template <typename T> class DynamicValueDataNum : public DynamicValueDataBase
{
	public:
		DynamicValueDataNum(T pv)
		{
			v = pv;
		}
		
		virtual DynamicValueDataBase *copy()
		{
			return new DynamicValueDataNum<T>(*this);
		}
		
		virtual DynamicValueDataBase *Op(char op, DynamicValueDataBase *rhs)
		{
			DynamicValueDataNum<T> *trhs = dynamic_cast<DynamicValueDataNum<T> *>(rhs);
			if (!trhs){
				throw FormatError("Left hand side and right hand side are of different types");
			}
			T nv;
			if (op == '+'){
				nv = v + trhs->v;
			}
			else if (op == '-'){
				nv = v - trhs->v;
			}
			else if (op == '*'){
				nv = v - trhs->v;
			}
			else if (op == '/'){
				nv = v - trhs->v;
			}
			else {
				throw FormatError("Oparator %c is not supported on numerical types");
			}
			return new DynamicValueDataNum<T>(nv);
		}
		
		virtual operator T()
		{
			return v;
		}
		
		virtual std::string ToString()
		{
			char buffer[1024];
			if (dynamic_cast<DynamicValueDataNum<int64_t> *>(this)){
				snprintf(buffer, 1024, "%" PRIi64, (int64_t)v);
			}
			else if (dynamic_cast<DynamicValueDataNum<double> *>(this)){
				snprintf(buffer, 1024, "%f", (double)v);
			}
			else {
				snprintf(buffer, 1024, "Unknown");
			}
			return buffer;
		}
		
	private:
		virtual const char *GetValueType()
		{
			return typeid(T).name();
		}
		
		T v;
};

class DynamicValueDataList : public DynamicValueDataBase
{
	public:
		DynamicValueDataList(std::vector<DynamicValue> pv)
		{
			v = pv;
		}
		
		virtual DynamicValueDataBase *copy()
		{
			return new DynamicValueDataList(*this);
		}
		
		virtual DynamicValueDataBase *Op(char op, DynamicValueDataBase *rhs)
		{
			DynamicValueDataList *trhs = dynamic_cast<DynamicValueDataList *>(rhs);
			if (!trhs){
				throw FormatError("Left hand side and right hand side are of different types");
			}
			if (op == '+'){
				std::vector<DynamicValue> l;
				l.insert(l.end(), v.begin(), v.end());
				l.insert(l.end(), trhs->v.begin(), trhs->v.end());
				return new DynamicValueDataList(l);
			}
			throw FormatError("Cannot perform operator %c on bools");
		}
		
		virtual operator std::vector<DynamicValue> &()
		{
			return v;
		}
		
		virtual std::string ToString()
		{
			std::string r = "[";
			for (std::vector<DynamicValue>::iterator it = v.begin(); it != v.end(); it++){
				if (r != "["){
					r += ", ";
				}
				r += it->ToString();
			}
			r += "]";
			
			return r;
		}
		
	private:
		virtual const char *GetValueType()
		{
			return "list";
		}
		
		std::vector<DynamicValue> v;
};

class DynamicValueDataMap : public DynamicValueDataBase
{
	public:
		DynamicValueDataMap(std::map<std::string, DynamicValue> pv)
		{
			v = pv;
		}
		
		virtual DynamicValueDataBase *copy()
		{
			return new DynamicValueDataMap(*this);
		}
		
		virtual DynamicValueDataBase *Op(char op, DynamicValueDataBase *rhs)
		{
			DynamicValueDataList *trhs = dynamic_cast<DynamicValueDataList *>(rhs);
			if (!trhs){
				throw FormatError("Left hand side and right hand side are of different types");
			}
			/*if (op == '+'){
				std::vector<DynamicValue> l;
				l.insert(l.end(), v.begin(), v.end());
				l.insert(l.end(), trhs->v.begin(), trhs->v.end());
				return new DynamicValueDataList(l);
			}*/
			throw FormatError("Cannot perform operator %c on bools");
		}
		
		virtual operator std::map<std::string, DynamicValue> &()
		{
			return v;
		}
		
		virtual std::string ToString()
		{
			std::string r = "";
			for (std::map<std::string, DynamicValue>::iterator it = v.begin(); it != v.end(); it++){
				r += std::string("\n");
				r += it->first + std::string(": ");
				r += it->second.ToString();
			}
			return Indent(r, 1);
		}
		
	private:
		virtual const char *GetValueType()
		{
			return "map";
		}
		
		std::map<std::string, DynamicValue> v;
};

class DynamicValueDataString : public DynamicValueDataBase
{
	public:
		DynamicValueDataString(std::string pv)
		{
			v = pv;
		}
		
		virtual DynamicValueDataBase *copy()
		{
			return new DynamicValueDataString(*this);
		}
		
		virtual DynamicValueDataBase *Op(char op, DynamicValueDataBase *rhs)
		{
			DynamicValueDataString *trhs = dynamic_cast<DynamicValueDataString *>(rhs);
			if (!trhs){
				throw FormatError("Left hand side and right hand side are of different types");
			}
			if (op == '+'){
				return new DynamicValueDataString(v + trhs->v);
			}
			throw FormatError("Cannot perform operator %c on strings");
		}
		
		virtual operator std::string()
		{
			return v;
		}
		
		virtual std::string ToString()
		{
			return v;
		}
		
	private:
		virtual const char *GetValueType()
		{
			return "string";
		}
		
		std::string v;
};

class DynamicValueDataTime : public DynamicValueDataBase
{
	public:
		DynamicValueDataTime(Time pv)
		{
			v = pv;
		}
		
		virtual DynamicValueDataBase *copy()
		{
			return new DynamicValueDataTime(*this);
		}
		
		virtual DynamicValueDataBase *Op(char op, DynamicValueDataBase *rhs)
		{
			DynamicValueDataString *trhs = dynamic_cast<DynamicValueDataString *>(rhs);
			if (!trhs){
				throw FormatError("Left hand side and right hand side are of different types");
			}
			throw FormatError("Cannot perform operator %c on times");
		}
		
		virtual operator Time()
		{
			return v;
		}
		
		virtual std::string ToString()
		{
			return "TIME";
		}
		
	private:
		virtual const char *GetValueType()
		{
			return "time";
		}
		
		Time v;
};

class DynamicValueDataEvent : public DynamicValueDataBase
{
	public:
		DynamicValueDataEvent(Event pv)
		{
			v = pv;
		}
		
		virtual DynamicValueDataBase *copy()
		{
			return new DynamicValueDataEvent(*this);
		}
		
		virtual DynamicValueDataBase *Op(char op, DynamicValueDataBase *rhs)
		{
			DynamicValueDataString *trhs = dynamic_cast<DynamicValueDataString *>(rhs);
			if (!trhs){
				throw FormatError("Left hand side and right hand side are of different types");
			}
			throw FormatError("Cannot perform operator %c on events");
		}
		
		virtual operator Event()
		{
			return v;
		}
		
		virtual std::string ToString()
		{
			return v.ToString();
		}
		
	private:
		virtual const char *GetValueType()
		{
			return "event";
		}
		
		Event v;
};

class DynamicValueDataBuffer : public DynamicValueDataBase
{
	public:
		DynamicValueDataBuffer(Buffer pv)
		{
			v = pv;
		}
		
		virtual DynamicValueDataBase *copy()
		{
			return new DynamicValueDataBuffer(*this);
		}
		
		virtual DynamicValueDataBase *Op(char op, DynamicValueDataBase *rhs)
		{
			DynamicValueDataString *trhs = dynamic_cast<DynamicValueDataString *>(rhs);
			if (!trhs){
				throw FormatError("Left hand side and right hand side are of different types");
			}
			throw FormatError("Cannot perform operator %c on times");
		}
		
		virtual operator Buffer()
		{
			return v;
		}
		
		virtual std::string ToString()
		{
			std::string r = "\n";
			for (unsigned int i = 0; i < v.GetLength(); i++){
				char buffer[4];
				unsigned int b = v[i];
				char c = ' ';
				if (i == v.GetLength() - 1){
					c = '\0';
				}
				else if (((i % 16) == 15) && (i)){
					c = '\n';
				}
				
				snprintf(buffer, 4, "%02X%c", b, c);
				r += buffer;
			}
			char buffer[19];
			snprintf(buffer, 19, "Length: %u", v.GetLength());
			return std::string(buffer) + Indent(r, 1);
		}
		
	private:
		virtual const char *GetValueType()
		{
			return "buffer";
		}
		
		Buffer v;
};


DynamicValue::DynamicValue()
{
	data = 0;
	ro = false;
}

DynamicValue::DynamicValue(bool v)
{
	data = new DynamicValueDataBool(v);
	ro = false;
}

DynamicValue::DynamicValue(int8_t v)
{
	data = new DynamicValueDataNum<int64_t>(v);
	ro = false;
}

DynamicValue::DynamicValue(uint8_t v)
{
	data = new DynamicValueDataNum<int64_t>(v);
	ro = false;
}

DynamicValue::DynamicValue(int16_t v)
{
	data = new DynamicValueDataNum<int64_t>(v);
	ro = false;
}

DynamicValue::DynamicValue(uint16_t v)
{
	data = new DynamicValueDataNum<int64_t>(v);
	ro = false;
}

DynamicValue::DynamicValue(int32_t v)
{
	data = new DynamicValueDataNum<int64_t>(v);
	ro = false;
}

DynamicValue::DynamicValue(uint32_t v)
{
	data = new DynamicValueDataNum<int64_t>(v);
	ro = false;
}

DynamicValue::DynamicValue(int64_t v)
{
	data = new DynamicValueDataNum<int64_t>(v);
	ro = false;
}

DynamicValue::DynamicValue(uint64_t v)
{
	data = new DynamicValueDataNum<int64_t>(v);
	ro = false;
}

DynamicValue::DynamicValue(void *v)
{
	data = new DynamicValueDataPointer(v);
	ro = false;
}

DynamicValue::DynamicValue(double v)
{
	data = new DynamicValueDataNum<double>(v);
	ro = false;
}

DynamicValue::DynamicValue(std::string v)
{
	data = new DynamicValueDataString(v);
	ro = false;
}

DynamicValue::DynamicValue(const char *v)
{
	data = new DynamicValueDataString(v);
	ro = false;
}

DynamicValue::DynamicValue(Time v)
{
	data = new DynamicValueDataTime(v);
	ro = false;
}

DynamicValue::DynamicValue(Event v)
{
	data = new DynamicValueDataEvent(v);
	ro = false;
}

DynamicValue::DynamicValue(Buffer v)
{
	data = new DynamicValueDataBuffer(v);
	ro = false;
}

DynamicValue::DynamicValue(std::vector<DynamicValue> v)
{
	data = new DynamicValueDataList(v);
	ro = false;
}

DynamicValue::DynamicValue(std::map<std::string, DynamicValue> v)
{
	data = new DynamicValueDataMap(v);
	ro = false;
}



/*DynamicValue(std::map<std::string, bool> v);
DynamicValue(std::map<std::string, int8_t> v);
DynamicValue(std::map<std::string, uint8_t> v);
DynamicValue(std::map<std::string, int16_t> v);
DynamicValue(std::map<std::string, uint16_t> v);
DynamicValue(std::map<std::string, int32_t> v);
DynamicValue(std::map<std::string, uint32_t> v);
DynamicValue(std::map<std::string, int64_t> v);
DynamicValue(std::map<std::string, uint64_t> v);
DynamicValue(std::map<std::string, const char *>v);
DynamicValue(std::map<std::string, void *>v);
DynamicValue(std::map<std::string, double> v);
DynamicValue(std::map<std::string, std::string> v);
DynamicValue(std::map<std::string, Time> v);
DynamicValue(std::map<std::string, Buffer> v);
DynamicValue(std::map<std::string, Event> v);*/


/*D: - It looks like we have to repeat this for all the ones we want to support
(and on conversion on the way out if we do that).
At least we get to handle all the annoying verbose stuff all in one place,
and have it reused*/
DynamicValue::DynamicValue(std::map<std::string, std::string> v)
{
	std::map<std::string, DynamicValue> dv;
	std::map<std::string, std::string>::iterator it;
	for (it = v.begin(); it != v.end(); it++){
		dv[it->first] = it->second;
	}
	
	data = new DynamicValueDataMap(dv);
	ro = false;
}






DynamicValue::DynamicValue(const DynamicValue &v)
{
	data = 0;
	if (v.data){
		data = v.data->copy();
	}
	ro = v.ro;
	format = v.format;
}

DynamicValue &DynamicValue::operator=(const DynamicValue &v)
{
	if (&v != this){
		DynamicValueDataBase *copydata = 0;
		if (v.data){
			copydata = v.data->copy();
		}
		
		if (data){
			delete data;
		}
		
		data = copydata;
		ro = v.ro;
		format = v.format;
	}
	return *this;
}

DynamicValue::~DynamicValue()
{
	if (data){
		delete data;
	}
}


DynamicValue DynamicValue::Op(char op, const DynamicValue &rhs) const
{
	DynamicValue v;
	if (data && rhs.data){
		v.data = data->Op(op, rhs.data);
	}
	return v;
}

std::string DynamicValue::ToString()
{
	if (format){
		return format.DoFormat(*this);
	}
	else if (data){
		return data->ToString();
	}
	else {
		return "None";
	}
}

void DynamicValue::SetRoKey()
{
	ro = true;
	//TODO: make this do the children!!!
	//TODO: Find out why Format() doesn't propogate through events/DynamicValues
}

void DynamicValue::Format(DynamicValueFormat pformat)
{
	format = pformat;
}

DynamicValue::operator bool()
{
	if (data){
		return *data;
	}
	throw FormatError("Cannot convert none to bool");
}

DynamicValue::operator int8_t()
{
	if (data){
		return (int64_t)*data;
	}
	throw FormatError("Cannot convert none to int");
}

DynamicValue::operator uint8_t()
{
	if (data){
		return (int64_t)*data;
	}
	throw FormatError("Cannot convert none to int");
}

DynamicValue::operator int16_t()
{
	if (data){
		return (int64_t)*data;
	}
	throw FormatError("Cannot convert none to int");
}

DynamicValue::operator uint16_t()
{
	if (data){
		return (int64_t)*data;
	}
	throw FormatError("Cannot convert none to int");
}

DynamicValue::operator int32_t()
{
	if (data){
		return (int64_t)*data;
	}
	throw FormatError("Cannot convert none to int");
}

DynamicValue::operator uint32_t()
{
	if (data){
		return (int64_t)*data;
	}
	throw FormatError("Cannot convert none to int");
}

DynamicValue::operator int64_t()
{
	if (data){
		return *data;
	}
	throw FormatError("Cannot convert none to int");
}

DynamicValue::operator uint64_t()
{
	if (data){
		return (int64_t)*data;
	}
	throw FormatError("Cannot convert none to int");
}

DynamicValue::operator void *()
{
	if (data){
		return *data;
	}
	throw FormatError("Cannot convert none to pointer");
}

DynamicValue::operator double()
{
	if (data){
		return *data;
	}
	throw FormatError("Cannot convert none to double");
}

DynamicValue::operator std::string()
{
	if (data){
		return *data;
	}
	throw FormatError("Cannot convert none to string");
}

DynamicValue::operator Time()
{
	if (data){
		return *data;
	}
	throw FormatError("Cannot convert none to time");
}

DynamicValue::operator Event()
{
	if (data){
		return *data;
	}
	throw FormatError("Cannot convert none to event");
}

DynamicValue::operator Buffer()
{
	if (data){
		return *data;
	}
	throw FormatError("Cannot convert none to buffer");
}

DynamicValue::operator std::vector<DynamicValue>()
{
	if (data){
		return *data;
	}
	throw FormatError("Cannot convert none to list");
}

DynamicValue::operator std::map<std::string, DynamicValue>()
{
	if (data){
		return *data;
	}
	throw FormatError("Cannot convert none to map");
}

DynamicValue &DynamicValue::operator[](unsigned int i)
{
	if (!data){
		throw FormatError("Cannot convert none to list getting item %u", i);
	}
	std::vector<DynamicValue> &l = *data;
	if (ro && (l.size() >= i)){
		throw FormatError("Cannot get key %u from list", i);
	}
	return l[i];
}

DynamicValue &DynamicValue::operator[](const char *key)
{
	if (!data){
		//throw FormatError("Cannot convert none to map");
		std::map<std::string, DynamicValue> l;
		data = new DynamicValueDataMap(l);
	}
	std::map<std::string, DynamicValue> &l = *data;
	if (ro && (!l.count(key))){
		throw FormatError("%s not in map", key);
	}
	return l[key];
}

bool DynamicValue::Has(std::string key)
{
	if (!data){
		throw FormatError("None cannot have keys (%s)\n", key.c_str());
	}
	std::map<std::string, DynamicValue> &l = *data;
	return l.count(key);
}

DynamicValue &DynamicValue::operator[](std::string key)
{
	if (!data){
		if (ro){
			throw FormatError("None cannot have key %s", key.c_str());
		}
		//throw FormatError("Cannot convert none to map");
		std::map<std::string, DynamicValue> l;
		data = new DynamicValueDataMap(l);
	}
	std::map<std::string, DynamicValue> &l = *data;
	if (ro && (!l.count(key))){
		throw FormatError("%s not in map", key.c_str());
	}
	return l[key];
}

std::vector<DynamicValue> &DynamicValue::GetVector()
{
	if (!data){
		throw FormatError("Cannot convert none to list");
	}
	return *data;
}

std::map<std::string, DynamicValue> &DynamicValue::GetMap()
{
	if (!data){
		throw FormatError("Cannot convert none to map");
	}
	return *data;
}
