/*
	Hyperion - A network anomaly discovery system
	Copyright (C) 2011-2012 Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "config.h"
#include "error.hpp"

#include "event.hpp"

#ifdef __GNUG__
#include <cstdlib>
#include <cxxabi.h>
#endif

#include "timer.hpp"

Event::Event()
{
	timestamp = Time::now();
	std::map<std::string, DynamicValue> l;
	root = DynamicValue(l);
}

Event::~Event()
{
}

Time Event::GetTimestamp()
{
	return timestamp;
}

std::string Event::GetFilter()
{
	std::string r;
	std::map<std::string, DynamicValue> l = root;
	for (std::map<std::string, DynamicValue>::iterator it = l.begin(); it != l.end(); it++){
		if (r != ""){
			r += ";";
		}
		r += it->first;
	}
	return r;
}

std::string Event::ToString()
{
	return root.ToString();
}

unsigned int Event::Count()
{
	std::map<std::string, DynamicValue> l = root;
	return l.size();
}

bool Event::Has(std::string name)
{
	std::map<std::string, DynamicValue> l = root;
	return l.count(name);
}

DynamicValue &Event::operator[](std::string key)
{
	return root[key];
}

EventSubscription::EventSubscription()
{
	refcount = 0;
}

EventSubscription::~EventSubscription()
{
}

void EventSubscription::IncRefCount()
{
	refcount++;
}

void EventSubscription::DecRefCount()
{
	refcount--;
	if (!refcount){
		delete this;
	}
}

EventManager::EventManager(Modules *pmodules)
{
	modules = pmodules;
}

EventManager::~EventManager()
{
	for (std::map<std::string, EventTable<uint32_t> *>::iterator it = inttables.begin(); it != inttables.end(); it++){
		delete it->second;
	}
	
	for (std::set<EventSubscription *>::iterator it = subscriptions.begin(); it != subscriptions.end(); it++){
		delete *it;
	}
}

void EventManager::Publish(Event event)
{
	queue.push_back(event);
	if (queue.size() == 1){
		while (queue.size()){
			event = queue.front();
			for (std::set<EventSubscription *>::iterator it = subscriptions.begin(); it != subscriptions.end(); it++){
				//This try/catch is essential in order to avoid event queue bugs
				//if something throws an error
				//TODO: Put the printfs() back to throws
				
				if (!unsubscriptions.count(*it)){
					//we haven't unsubscribed from this subscription
					try
					{
						(*it)->HandleEvent(event);
					}
					catch (Error e)
					{
						e.OutputMessage();
					}
				}
			}
			queue.pop_front();
			
			while (unsubscriptions.size()){ //remove any subscriptions that have
				//been unsubscribed
				subscriptions.erase(*unsubscriptions.begin()); //remove the subscription
				unsubscriptions.erase(*unsubscriptions.begin()); //remobe the unsubscription
			}
		}
	}
}

void EventManager::Subscribe(EventSubscription *subscription)
{
	subscriptions.insert(subscription);
}

void EventManager::Unsubscribe(EventSubscription *subscription)
{
	unsubscriptions.insert(subscription); //add to the list of unsubscriptions
	delete subscription; //the subscription is no longer valid, so delete it
}

EventTable<uint32_t> *EventManager::GetIntEventTable(std::string name)
{
	if (!inttables.count(name)){
		inttables[name] = new EventTable<uint32_t>(name, modules, this);
	}
	return inttables[name];
}
