/*
	Hyperion - A network anomaly discovery system
	Copyright (C) 2011 Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "config.h"
#include "error.hpp"

#include <cstdio>
#include <typeinfo>

#include "event.hpp"
#include "module.hpp"

class EthernetModule : public Module
{
	public:
		EthernetModule(Modules *modules, EventManager *manager, Configuration *configuration, Database *database) :
			Module("Ethernet", modules, manager, configuration, database)
		{
			GetManager()->Subscribe(GetCallbackNameEventSubscription<EthernetModule>
				("received:ethernet", this, &EthernetModule::OnReceive));
			
			GetManager()->Subscribe(GetCallbackNameEventSubscription<EthernetModule>
				("send:ethernet", this, &EthernetModule::OnSend));
			
			GetManager()->GetIntEventTable("ethernet")->Add("received:ipv4", 0x0800);
			GetManager()->GetIntEventTable("ethernet")->Add("received:arp", 0x0806);
		}
		
	private:
		void OnSend(Event event)
		{
			Buffer src = event["send:ethernet"]["source"];
			Buffer dst = event["send:ethernet"]["destination"];
			Buffer proto(2);
			proto.SetUInt16N(0, event["send:ethernet"]["protocol"]);
			
			//TODO: See if we actually need all these on new lines
			Buffer b = dst;
			b = b + src;
			b = b + proto;
			b = b + event["send:ethernet"]["payload"];
			
			std::string name = event["send:ethernet"]["next-name"];
			Event e = event["send:ethernet"]["next"];
			e[name]["payload"] = b;
			e[name]["previous"] = event;
			GetManager()->Publish(e);
		}
		
		void OnReceive(Event event)
		{
			Buffer b = event["received:ethernet"]["data"];
			Event e;
			if (b.GetLength() < 14){
				e["anomaly:received:ethernet"]["data"] = b;
				GetManager()->Publish(e);
			}
			else {
				uint16_t ethertype = b.GetUInt16N(12);
				Buffer payload(b, 14);
				
				char strether[7];
				snprintf(strether, 7, "0x%04Xh", ethertype);
				
				e["processed:ethernet"]["payload"] = payload;
				e["processed:ethernet"]["ethertype"] = ethertype;
				e["processed:ethernet"]["source"] = Buffer(&b[6], 6);
				e["processed:ethernet"]["destination"] = Buffer(&b[0], 6);
				e["processed:ethernet"]["ethertype"].Format(FormatHex());
				e["processed:ethernet"]["source"].Format(FormatMac());
				e["processed:ethernet"]["destination"].Format(FormatMac());
				GetManager()->Publish(e);
				
				GetManager()->GetIntEventTable("ethernet")->Publish(ethertype, payload);
			}
		}
		
	std::string GetAddress(const unsigned char addr[6]) const
	{
		char buffer[24];
		snprintf(buffer, 24, "%02X:%02X:%02X:%02X:%02X:%02X",
			addr[0], addr[1], addr[2], addr[3], addr[4], addr[5]);
		return std::string(buffer);
	}
};

INIT_MODULE(EthernetModule);
