/*
	Hyperion - A network anomaly discovery system
	Copyright (C) 2012 Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "config.h"
#include "error.hpp"

#include <algorithm>
#include <list>
#include <map>

#include <pcrecpp.h>

#include "event.hpp"
#include "fsm.hpp"
#include "module.hpp"

//TODO: Make an abstract stream so that we don't have to have TCP here

//NOTE: I could probalby have implemented this using flex and yacc, but this
//	may be better integrated with the event system

//NOTE: this module assumes stuff is sniffed in the order it's sent/received

class HttpModule;

//HTTP specifies that resonses must be in the same order as the pipelined requests
//This means a simple list will do
class HttpRequest
{
	public:
		std::string method;
		std::string resource;
		std::map<std::string, std::string> headers;
		std::map<std::string, std::string> headers_lower;
};

class HttpData
{
	public:
		void RemoveConnection();
		void PublishEventToParent(Event e);
		
		void PublishEvent(bool server, std::string name, DynamicValue v = DynamicValue(std::map<std::string, DynamicValue>()))
		{
			Event e;
			e[name]["id"] = id;
			e[name]["client address"] = client_event["client address"];
			e[name]["server address"] = client_event["server address"];
			e[name]["address type"] = client_event["address type"];
			e[name]["client port"] = client_event["client port"];
			e[name]["server port"] = client_event["server port"];
			
			if (server){
				if (requests.size()){
					e[name]["method"] = requests.front().method;
					e[name]["resource"] = requests.front().resource;
					e[name]["client headers"] = requests.front().headers;
					e[name]["client lowercase headers"] = requests.front().headers_lower;
					
					//if we can generate a URL, do so
					if (requests.front().headers_lower.count("host")){
						e[name]["url"] = std::string("http://") +
							requests.front().headers_lower["host"] +
							requests.front().resource;
					}
				}
				
				e[name]["status code"] = server_status_int;
				e[name]["status string"] = server_status_string;
				e[name]["server headers"] = server_headers;
				e[name]["server lowercase headers"] = server_headers_lower;
			}
			else {
				if (requests.size()){
					e[name]["method"] = requests.back().method;
					e[name]["resource"] = requests.back().resource;
					e[name]["client headers"] = requests.back().headers;
					e[name]["client lowercase headers"] = requests.back().headers_lower;
					
					//if we can generate a URL, do so
					if (requests.back().headers_lower.count("host")){
						e[name]["url"] = std::string("http://") +
							requests.back().headers_lower["host"] +
							requests.back().resource;
					}
				}
			}
			
			std::map<std::string, DynamicValue> m = v;
			for (std::map<std::string, DynamicValue>::iterator it = m.begin(); it != m.end(); it++){
				e[it->first] = it->second;
			}
		}
		
		HttpData(uint64_t pid, HttpModule *phttpmodule)
		{
			id = pid;
			httpmodule = phttpmodule;
			ClientReset();
			ServerReset();
		}
		
		void ClientReset()
		{
			client_content_read = 0;
			client_content_length = 0;
			client_data = "";
		}
		
		void ServerReset()
		{
			server_data = "";
		}
		
		
		std::list<HttpRequest> requests;
		
		//client stuff
		uint64_t client_content_read;
		uint64_t client_content_length;
		std::string client_data;
		Event client_event;
		
		//server stuff
		uint64_t server_content_read;
		uint64_t server_content_length;
		std::string server_data;
		Event server_event;
		int server_status_int;
		std::string server_status_string;
		std::map<std::string, std::string> server_headers_lower;
		std::map<std::string, std::string> server_headers;
		
	private:
		uint64_t id;
		HttpModule *httpmodule;
};

//returns true if we should move on, false if not
bool HttpConnectionFsmNodeProcessHeaders(
	bool server,
	HttpData *&data,
	std::string &stream_data,
	std::map<std::string, std::string> &headers,
	std::map<std::string, std::string> &headers_lower,
	uint64_t &content_length)
{
	//each iteration of the loop processes a line
	while (stream_data.find_last_of('\n') != std::string::npos){
		//we have at least one full line
		
		//pop off the first line
		size_t p = stream_data.find_first_of('\n');
		std::string line = stream_data.substr(0, p);
		stream_data = stream_data.substr(p);
		p = line.find_last_of('\r');
		line = line.substr(0, p);
		
		//check to see if the first character is a newline (which would
		//imply end of headers)
		if ((line[0] == '\n') || (line[0] == '\r')){
			//get lower case header keys
			for (std::map<std::string, std::string>::iterator it = headers.begin();
				it != headers.end(); it++){
				std::string key = it->first;
				std::transform (key.begin(), key.end(), key.begin(), tolower);
				headers_lower[key] = it->second;
			}
			
			//get the content length
			if (headers_lower.count("content-length")){
				content_length = std::atoll(headers_lower["content-length"].c_str());
			}
			//else it was already set to zero in HttpData::Reset()
			
			return true;
		}
		
		//process the line
		p = line.find_first_of(':');
		std::string key = line.substr(0, p);
		std::string value = "";
		
		if (p != std::string::npos){
			//TODO: remove all leading/trailing whitespace
			line.substr(p);
			if ((value[0] == ' ') || (value[0] == '\t')){
				//remove the first whitespace character from the start
				value = value.substr(1, std::string::npos);
			}
		}
		
		if (headers.count(key)){
			DynamicValue v;
			v["header"] = key;
			v["existing value"] = headers[key];
			v["new value"] = value;
			data->PublishEvent(server, "anomaly:received:http:redefined header", v);
			data->RemoveConnection();
			return false;
		}
		headers[key] = value;
	}
	
	//TODO: Make less arbitrary
	if (stream_data.size() > 4096){
		data->PublishEvent(server, "anomaly:received:http:header too long");
		data->RemoveConnection();
		return false;
	}
	
	return false;
}

bool HttpConnectionFsmNodeProcessData(
	HttpData *&data,
	std::string &stream_data,
	uint64_t &content_length,
	uint64_t &content_read)
{
	if (content_length > content_read){
		//we have data to skip over - so skip over it:p
		uint64_t remaining = content_length - content_read;
		if (stream_data.size() <= remaining){
			content_read += stream_data.size(); //record that we skipped over this
			stream_data = "";
		}
		else {
			stream_data = stream_data.substr(remaining, std::string::npos);
			content_read += remaining;
		}
	}
	
	if (content_length == content_read){
		//no more content, so we must be at the end of the request
		return true;
	}
	
	return false;
}

class HttpConnectionFsmNodeServerBase  : public FsmNode<Event, HttpData *>
{
	private:
		virtual FsmNode<Event, HttpData *> *Process(HttpData *&data) = 0;
		
		virtual FsmNode<Event, HttpData *> *ProcessTransitionIn(HttpData *&data)
		{
			return Process(data);
		}
		
		virtual FsmNode<Event, HttpData *> *ProcessCharacter(Event event, HttpData *&data)
		{
			data->server_event = event;
			Buffer d = event["processed:tcp:server data"]["data"];
			data->server_data.append((const char *)d.GetBuffer(), d.GetLength());
			
			return Process(data);
		}
};

class HttpConnectionFsmNodeServerData : public HttpConnectionFsmNodeServerBase
{
	public:
		HttpConnectionFsmNodeServerData(FsmNode<Event, HttpData *> *start) :
			next(start){}
		
	private:
		virtual FsmNode<Event, HttpData *> *Process(HttpData *&data)
		{
			if (HttpConnectionFsmNodeProcessData(
				data,
				data->server_data,
				data->server_content_length,
				data->server_content_read)){
				data->ServerReset(); //it's easier to have this here
				
				data->PublishEvent(true, "processed:http:response");
				data->requests.pop_front();
				
				return next;
			}
			return this;
		}
		
		FsmNode<Event, HttpData *> *next;
};

class HttpConnectionFsmNodeServerHeaders : public HttpConnectionFsmNodeServerBase
{
	public:
		HttpConnectionFsmNodeServerHeaders(FsmNode<Event, HttpData *> *start) :
			next(start){}
		
	private:
		virtual FsmNode<Event, HttpData *> *Process(HttpData *&data)
		{
			if (HttpConnectionFsmNodeProcessHeaders(
				true,
				data,
				data->server_data,
				data->server_headers,
				data->server_headers_lower,
				data->server_content_length)){
				return &next;
			}
			return this;
		}
		
		HttpConnectionFsmNodeServerData next;
};

class HttpConnectionFsmNodeServerStart : public HttpConnectionFsmNodeServerBase
{
	public:
		HttpConnectionFsmNodeServerStart() :
			next(this){}
		
	private:
		virtual FsmNode<Event, HttpData *> *Process(HttpData *&data)
		{
			size_t p = data->server_data.find_first_of('\n');
			if (p != std::string::npos){ //we found the first line
				//pop the first line
				std::string line = data->server_data.substr(0, p);
				data->server_data.substr(p + 1);
				
				//remove the \r from the first line if there is one
				p = line.find_first_of('\r');
				if (p != std::string::npos){
					line = line.substr(0, p);
				}
				
				//pop the HTTP version
				p = line.find_first_of(' ');
				if (p == std::string::npos){
					data->PublishEvent(true, "anomaly:received:http:invalid server response");
					data->RemoveConnection();
					return this;
				}
				line = line.substr(p + 1);
				
				//pop the status code off
				p = line.find_first_of(' ');
				if (p == std::string::npos){
					data->PublishEvent(true, "anomaly:received:http:invalid server response");
					data->RemoveConnection();
					return this;
				}
				std::string status_int = line.substr(0, p);
				line = line.substr(p + 1);
				data->server_status_int = atoi(status_int.c_str());
				
				//The remainder of the line is the string of the status
				data->server_status_string = line;
				
				return &next;
			}
			else if (data->server_data.size() > 4096){
				//TODO: ARBITRARY - FIX/ALLOW CUSTOMIZATION
				//(or take it from the HTTP RFC's spec for maximum resource length)
				data->PublishEvent(true, "anomaly:received:http:server first line too long");
				data->RemoveConnection();
				return this; //return this so we don't write
			}
			
			return this;
		}
		
		HttpConnectionFsmNodeServerHeaders next;
};

//base class to abstract the common functionality in the Process methods
class HttpConnectionFsmNodeClientBase  : public FsmNode<Event, HttpData *>
{
	private:
		virtual FsmNode<Event, HttpData *> *Process(HttpData *&data) = 0;
		
		virtual FsmNode<Event, HttpData *> *ProcessTransitionIn(HttpData *&data)
		{
			return Process(data);
		}
		
		virtual FsmNode<Event, HttpData *> *ProcessCharacter(Event event, HttpData *&data)
		{
			data->client_event = event;
			Buffer d = event["processed:tcp:client data"]["data"];
			data->client_data.append((const char *)d.GetBuffer(), d.GetLength());
			
			return Process(data);
		}
};

class HttpConnectionFsmNodeClientData : public HttpConnectionFsmNodeClientBase
{
	public:
		HttpConnectionFsmNodeClientData(FsmNode<Event, HttpData *> *start) :
			next(start){}
		
	private:
		virtual FsmNode<Event, HttpData *> *Process(HttpData *&data)
		{
			if (HttpConnectionFsmNodeProcessData(
				data,
				data->client_data,
				data->client_content_length,
				data->client_content_read)){
				data->ClientReset(); //it's easier to have this here
				return next;
			}
			return this;
		}
		
		FsmNode<Event, HttpData *> *next;
};

class HttpConnectionFsmNodeClientHeaders : public HttpConnectionFsmNodeClientBase
{
	public:
		HttpConnectionFsmNodeClientHeaders(FsmNode<Event, HttpData *> *start) :
			next(start){}
		
	private:
		virtual FsmNode<Event, HttpData *> *Process(HttpData *&data)
		{
			if (HttpConnectionFsmNodeProcessHeaders(
				false,
				data,
				data->client_data,
				data->requests.back().headers,
				data->requests.back().headers_lower,
				data->client_content_length)){
				return &next;
			}
			return this;
		}
		
		HttpConnectionFsmNodeClientData next;
};

//this is designed to process the request line only (e.g: "GET / HTTP/1.1")
class HttpConnectionFsmNodeClientStart : public HttpConnectionFsmNodeClientBase
{
	public:
		HttpConnectionFsmNodeClientStart() :
			next(this){}
	
	private:
		virtual FsmNode<Event, HttpData *> *Process(HttpData *&data)
		{
			if (http.FullMatch(data->client_data)){ //we got a good match!
				HttpRequest request;
				
				//Get the first line
				size_t p = data->client_data.find_first_of('\n');
				std::string line = data->client_data.substr(0, p);
				data->client_data = data->client_data.substr(p);
				
				//Get the method, and remove it from the client data
				p = line.find_first_of(' ');
				request.method = line.substr(0, p);
				line = line.substr(p + 1);
				
				//Get the resource, and remove it from the client data
				p = line.find_last_of(' ');
				request.resource = line.substr(0, p);
				
				//add the request to the queue
				data->requests.push_back(request);
				
				return &next;
			}
			else if ((data->client_data.size() > 4096) || (nothttp.FullMatch(data->client_data))){
				//TODO: ARBITRARY - FIX/ALLOW CUSTOMIZATION
				//(or take it from the HTTP RFC's spec for maximum resource length)
				//Note: this is not necessarily an anomaly (may just be > 4096 because it's not HTTP)
				data->RemoveConnection();
				//printf("********MISMATCH********\n%s\n********\n\n", data->client_data.c_str());
				return this; //return this so we don't write
			}
			return this;
		}
		
		HttpConnectionFsmNodeClientHeaders next; //This is a bit odd if you think about it
			//in the traditional OOP sense, but it is a convenient thing to do
		
		static pcrecpp::RE http;
		static pcrecpp::RE nothttp;
};

//perhaps modify this so it matches specifically against GET|POST|...
//Disadvantage would be that this would prevent unknown method as being an anomaly
pcrecpp::RE HttpConnectionFsmNodeClientStart::http("[A-Z]+ [^ ]+ HTTP/[0-9.]+\r?\n.*");
pcrecpp::RE HttpConnectionFsmNodeClientStart::nothttp(".*\n.*"); //designed to be applied after http

class HttpConnection
{
	public:
		HttpConnection(uint64_t id, HttpModule *pparent, EventManager *pmanager) :
			httpdata(id, pparent),
			clientfsm(&httpdata, &client_start_node),
			serverfsm(&httpdata, &server_start_node)
		{
			manager = pmanager;
			parent = pparent;
			
			server_data_subscription = GetCallbackNameValueEventSubscription <HttpConnection, uint64_t>
				("processed:tcp:server data", "id", id, this, &HttpConnection::OnServerData);
			
			client_data_subscription = GetCallbackNameValueEventSubscription <HttpConnection, uint64_t>
				("processed:tcp:client data", "id", id, this, &HttpConnection::OnClientData);
			
			manager->Subscribe(server_data_subscription);
			manager->Subscribe(client_data_subscription);
		}
		
		~HttpConnection()
		{
			manager->Unsubscribe(server_data_subscription);
			manager->Unsubscribe(client_data_subscription);
		}
		
	private:
		void OnClientData(Event event)
		{
			clientfsm.ProcessCharacter(event);
		}
		
		void OnServerData(Event event)
		{
			serverfsm.ProcessCharacter(event);
		}
		
		HttpData httpdata;
		//WARNING: the nodes MUST be declared before the FSMs, otherwise the
		//	polymorphic transition in method will break
		HttpConnectionFsmNodeClientStart client_start_node;
		HttpConnectionFsmNodeServerStart server_start_node;
		Fsm<Event, HttpData *> clientfsm;
		Fsm<Event, HttpData *> serverfsm;
		
		EventManager *manager;
		HttpModule *parent;
		
		EventSubscription *server_data_subscription;
		EventSubscription *client_data_subscription;
};

class HttpModule : public Module
{
	public:
		HttpModule(Modules *modules, EventManager *manager, Configuration *configuration, Database *database) :
			Module("HTTP", modules, manager, configuration, database)
		{
			GetManager()->Subscribe(GetCallbackNameEventSubscription<HttpModule>
				("processed:tcp:new", this, &HttpModule::OnNew));
			
		}
		
		void RemoveConnection(uint64_t id)
		{
			GetManager()->Unsubscribe(close_subscriptions[id]);
			close_subscriptions.erase(id);
			delete connections[id];
			connections.erase(id);
		}
		
		void PublishEvent(Event e)
		{
			GetManager()->Publish(e);
		}
		
	private:
		void OnNew(Event event)
		{
			Event e;
			uint64_t id = event["processed:tcp:new"]["id"];
			
			connections[id] = new HttpConnection(id, this, GetManager());
			
			close_subscriptions[id] = GetCallbackNameValueEventSubscription <HttpModule, uint64_t>
				("processed:tcp:close", "id", id, this, &HttpModule::OnClose);
			
			GetManager()->Subscribe(close_subscriptions[id]);
		}
		
		void OnClose(Event event)
		{
			uint64_t id = event["processed:tcp:close"]["id"];
			RemoveConnection(id);
		}
		
		std::map<uint64_t, HttpConnection *> connections;
		std::map<uint64_t, EventSubscription *> close_subscriptions;
};

void HttpData::RemoveConnection()
{
	httpmodule->RemoveConnection(id);
}

void HttpData::PublishEventToParent(Event e)
{
	httpmodule->PublishEvent(e);
}

INIT_MODULE(HttpModule);
