/*
	Hyperion - A network anomaly discovery system
	Copyright (C) 2012 Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "config.h"
#include "error.hpp"

#include <map>
#include <set>

#include "event.hpp"
#include "module.hpp"
#include "structures.hpp"

class PortscanAddressInfo
{
	public:
		PortscanAddressInfo()
		{
			reported = false;
		}
		
		void SetReceived(uint16_t port)
		{
			if (!solicited.count(port)){
				unsolicited.insert(port);
			}
		}
		
		void SetSent(uint16_t port)
		{
			solicited.insert(port);
		}
		
		unsigned int GetUnsolicited()
		{
			return unsolicited.size();
		}
		
		bool Report()
		{
			if (reported){
				return false;
			}
			else {
				reported = true;
				return true;
			}
		}
		
	private:
		std::set<uint16_t> solicited;
		std::set<uint16_t> unsolicited;
		
		bool reported;
		
		//TODO: Have some mechanism not to rereport within a specified time
};

class PortscanModule : public Module
{
	public:
		PortscanModule(Modules *modules, EventManager *manager, Configuration *configuration, Database *database) :
			Module("Portscan", modules, manager, configuration, database)
		{
			GetManager()->Subscribe(GetCallbackNameEventSubscription<PortscanModule>
				("processed:tcp:packet", this, &PortscanModule::OnReceiveTcp));
			GetManager()->Subscribe(GetCallbackNameEventSubscription<PortscanModule>
				("processed:udp", this, &PortscanModule::OnReceiveUdp));
			
			//TODO: Do this with a database
			/*Let's have a database of remote connections*/
			/*GetDatabase()("CREATE TABLE IF NOT EXISTS portscan_receives\
				(id INTEGER PRIMARY KEY,\
				src BINARY NOT NULL,\
				netname VARCHAR(64) NOT NULL,\
				dport SMALLINT NOT NULL,\
				timestamp INTEGER NOT NULL);")();*/
		}
		
	private:
		void OnReceiveTcp(Event event)
		{
			OnReceive(event["processed:tcp:packet"], event);
		}
		
		void OnReceiveUdp(Event event)
		{
			OnReceive(event["processed:udp"], event);
		}
		
		void OnReceive(DynamicValue value, Event event)
		{
			uint16_t sport = value["source port"];
			uint16_t dport = value["destination port"];
			DynamicValue src = value["source address"];
			DynamicValue dst = value["destination address"];
			ComparisonPair<Buffer, Buffer> srcdst(src, dst);
			ComparisonPair<Buffer, Buffer> dstsrc(dst, src);
			addrs[dstsrc].SetSent(sport);
			addrs[srcdst].SetReceived(dport);
			
			unsigned int threshold = GetConfiguration()["portscan"]["threshold"];
			bool multirep = GetConfiguration()["portscan"]["repeat"];
			unsigned int count = addrs[srcdst].GetUnsolicited();
			if ((count > threshold) && (multirep || addrs[srcdst].Report())){
				DynamicValue v;
				v["source"] = src;
				v["destination"] = dst;
				v["count"] = count;
				
				Event e;
				e["anomaly:portscan:too many unsolicited"] = v;
				GetManager()->Publish(e);
			}
		}
		
		//Let's consider this to be remote, local
		std::map<ComparisonPair<Buffer, Buffer>, PortscanAddressInfo> addrs;
};

INIT_MODULE(PortscanModule);
 
