/*
	Hyperion - A network anomaly discovery system
	Copyright (C) 2012 Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "config.h"
#include "error.hpp"

#define __STDC_FORMAT_MACROS
#include "inttypes.h"
#include "stdint.h"
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <unistd.h>

#include "event.hpp"
#include "module.hpp"
#include "time.hpp"

class ResourceMonitorModule : public Module
{
	public:
		ResourceMonitorModule(Modules *modules, EventManager *manager, Configuration *configuration, Database *database) :
			Module("Resource Monitor", modules, manager, configuration, database)
		{
			GetManager()->Subscribe(GetCallbackNameEventSubscription<ResourceMonitorModule>
				("sniff", this, &ResourceMonitorModule::OnReceive));
			GetManager()->Subscribe(GetCallbackNameEventSubscription<ResourceMonitorModule>
				("send:sniff", this, &ResourceMonitorModule::OnSend));
			
			std::string filename = GetConfiguration()["resmon"]["filename"];
			bool append = GetConfiguration()["resmon"]["append"];
			if (append){
				f = fopen(filename.c_str(), "a");
			}
			else {
				f = fopen(filename.c_str(), "w");
			}
			
			if (!f){
				throw FormatError("Could not open file %s", filename.c_str());
			}
			
			t = Time::now();
		}
		
	private:
		void OnReceive(Event event)
		{
			Buffer b = event["sniff"]["data"];
			
			rusage usage;
			if (getrusage(RUSAGE_SELF, &usage)){
				throw FormatError("Could not get resource usage: %s", errstr);
			}
			
			WriteUsage(usage, "receive", b.GetLength());
		}
		
		void OnSend(Event event)
		{
			Buffer b = event["send:sniff"]["payload"];
			
			rusage usage;
			if (getrusage(RUSAGE_SELF, &usage)){
				throw FormatError("Could not get resource usage: %s", errstr);
			}
			
			WriteUsage(usage, "send", b.GetLength());
		}
		
		void WriteUsage(rusage &usage, const char *type, unsigned int length)
		{
			uint64_t size = 0;
			uint64_t resident = 0;
			uint64_t share = 0;
			uint64_t text = 0;
			uint64_t lib = 0;
			uint64_t data = 0;
			uint64_t dt = 0;
			
			GetProcPidStatm(size, resident, share, text, lib, data, dt);
			size *= GetPageSize();
			resident *= GetPageSize();
			share *= GetPageSize();
			text *= GetPageSize();
			lib *= GetPageSize();
			data *= GetPageSize();
			dt *= GetPageSize();
			
			//Some of these descriptions (comments) are copied from die.net
			fprintf(f, "%s\t%u\t%f\t%f\t%f\t%li\t%li\t%li\t%li\t%li\t%li\t%li\t%li\t%li\t%li\t%li\t%li\t%li\t%li\t"
				"%"PRIu64"\t%"PRIu64"\t%"PRIu64"\t%"PRIu64"\t%"PRIu64"\t%"PRIu64"\t%"PRIu64"\n",
				type, /* type */
				length, /* size transferred */
				(double)(Time::now() - t), /* time since program start */
				(double)Time(usage.ru_utime), /* user CPU time used */
				(double)Time(usage.ru_stime), /* system CPU time used */
				usage.ru_maxrss, /* maximum resident set size */
				usage.ru_ixrss, /* integral shared memory size */
				usage.ru_idrss, /* integral unshared data size */
				usage.ru_isrss, /* integral unshared stack size */
				usage.ru_minflt, /* page reclaims (soft page faults) */
				usage.ru_majflt, /* page faults (hard page faults) */
				usage.ru_nswap, /* swaps */
				usage.ru_inblock, /* block input operations */
				usage.ru_oublock, /* block output operations */
				usage.ru_msgsnd, /* IPC messages sent */
				usage.ru_msgrcv, /* IPC messages received */
				usage.ru_nsignals, /* signals received */
				usage.ru_nvcsw, /* voluntary context switches */
				usage.ru_nivcsw, /* involuntary context switches */
				
				size, /* total program size */
				resident, /* resident set size */
				share, /* shared pages */
				text, /* text (code) */
				lib, /* library (unused in Linux 2.6) */
				data, /* data + stack */
				dt /* dirty pages (unused in Linux 2.6) */
   			);
		}
		
		void GetProcPidStatm(uint64_t &size, uint64_t &resident, uint64_t &share,
			uint64_t &text, uint64_t &lib, uint64_t &data, uint64_t &dt)
		{
			int pid = getpid();
			char filename[1024];
			snprintf(filename, 1024, "/proc/%i/statm", pid);
			FILE *statm = fopen(filename, "r");
			if (!statm){
				throw FormatError("Could not open %s", filename);
			}
			if (fscanf(statm, "%"PRIu64" %"PRIu64" %"PRIu64" %"PRIu64" %"PRIu64" %"PRIu64" %"PRIu64"\n",
				&size, &resident, &share, &text, &lib, &data, &dt) != 7){
				throw FormatError("Could not read statm");
			}
			fclose(statm);
		}
		
		uint64_t GetPageSize()
		{
			return 4096;
		}
		
		Time t;
		FILE *f;
};

INIT_MODULE(ResourceMonitorModule);
 
 
