/*
	Hyperion - A network anomaly discovery system
	Copyright (C) 2011-2012 Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "config.h"
#include "error.hpp"

#include <map>

#include "assemblybuffer.hpp"
#include "event.hpp"
#include "fsm.hpp"
#include "module.hpp"

//TODO: implement limited window, etc, to make sure no memory leak attacks exist (SECURITY)

class TcpModule;
class TcpPacket
{
	public:
		TcpPacket(TcpModule &pparent, DynamicValue psrc, DynamicValue pdst, std::string pnetname,
			Buffer ppayload, DynamicValue pev, bool pclientorigin,
			uint16_t psport, uint16_t pdport, uint32_t psequence, uint32_t packnowledgement, uint16_t pwindow, uint16_t purgent,
			bool pfin, bool psyn, bool prst, bool ppsh, bool pack, bool purg) :
			parent(pparent),
			src(psrc),
			dst(pdst),
			netname(pnetname),
			payload(ppayload),
			ev(pev),
			clientorigin(pclientorigin),
			sport(psport),
			dport(pdport),
			sequence(psequence),
			acknowledgement(packnowledgement),
			window(pwindow),
			urgent(purgent),
			fin(pfin),
			syn(psyn),
			rst(prst),
			psh(ppsh),
			ack(pack),
			urg(purg){}
			
		TcpModule &parent;
		DynamicValue src;
		DynamicValue dst;
		std::string netname;
		Buffer payload;
		DynamicValue ev;
		bool clientorigin;
		uint16_t sport;
		uint16_t dport;
		uint32_t sequence;
		uint32_t acknowledgement;
		uint16_t window;
		uint16_t urgent;
		bool fin;
		bool syn;
		bool rst;
		bool psh;
		bool ack;
		bool urg;
};

class TcpConnectionDataStream
{
	public:
		TcpConnectionDataStream()
		{
			initial_sequence = 0;
			read = 0;
		}
		
		AssemblyBuffer buffer;
		uint64_t read;
		uint32_t initial_sequence;
};

class TcpConnectionData
{
	public:
		uint64_t id;
		
		DynamicValue serveraddr;
		DynamicValue clientaddr;
		DynamicValue serverport;
		DynamicValue clientport;
		DynamicValue inetname;
		
		TcpConnectionDataStream client_stream;
		TcpConnectionDataStream server_stream;
};

class TcpConnectionFsmNode : public FsmNode<TcpPacket &, TcpConnectionData>
{
	protected:
		void AddToBuffer(Buffer buffer, TcpConnectionDataStream &stream, TcpPacket &packet, TcpConnectionData &data)
		{
			if (!buffer.GetLength()){
				return;
			}
			
			//TODO: Find the RFC, and see how it says this should be done
			uint64_t sequence = packet.sequence;
			uint64_t readsequence = (uint64_t)stream.initial_sequence + stream.read;
			uint64_t readsequence_mod = readsequence & 0xFFFFFFFF;
			if (sequence < readsequence_mod){
				sequence += 0xFFFFFFFF;
				sequence += 1; //Meh:p - less effort than finding the correct extension:p
			}
			
			uint64_t offset = sequence - readsequence_mod;
			
			printf("Adding %u bytes at %lu\n", buffer.GetLength(), offset);
			
			stream.buffer.AddBuffer(offset, buffer);
			
			DynamicValue evi;
			if (stream.buffer.PopMaxLength()){
				Buffer payload = stream.buffer.PopMax();
				evi["id"] = data.id;
				evi["server address"] = data.serveraddr;
				evi["client address"] = data.clientaddr;
				evi["server port"] = data.serverport;
				evi["client port"] = data.clientport;
				evi["data"] = payload;
				stream.read += payload.GetLength();
				
				if (packet.clientorigin){
					PublishEvent("processed:tcp:client data", evi, packet.parent);
				}
				else {
					PublishEvent("processed:tcp:server data", evi, packet.parent);
				}
			}
		}
		
		void PublishEvent(std::string name, DynamicValue evi, TcpModule &module);
};

//No packets processed
class TcpConnectionFsmNodeInit1 : public TcpConnectionFsmNode
{
	public:
		TcpConnectionFsmNodeInit1(FsmNode<TcpPacket &, TcpConnectionData> *pinit2) :
			init2(pinit2){}
		
	private:
		//Expect a SYN packet
		virtual FsmNode<TcpPacket &, TcpConnectionData> *ProcessCharacter(TcpPacket &packet, TcpConnectionData &data)
		{
			//This should be first because it's intercepting a TCP stream half way through
			//For the same reason, it shouldn't emit an anomaly (TODO: at least not without configuration)
			if (packet.fin || !packet.syn || packet.rst || packet.psh || packet.ack || packet.urg){
				//parent.PublishEvent("anomaly:received:tcp:initial packet not syn", ev);
				return 0;
			}
			
			if (!packet.clientorigin){
				PublishEvent("anomaly:received:tcp:syn from server", packet.ev, packet.parent);
				return 0;
			}
			
			if (packet.payload.GetLength()){
				PublishEvent("anomaly:received:tcp:initial packet has payload", packet.ev, packet.parent);
				return 0;
			}
			
			data.id = cid;
			cid++;
			
			DynamicValue evi;
			evi["id"] = data.id;
			evi["client address"] = packet.src;
			evi["server address"] = packet.dst;
			evi["address type"] = packet.netname;
			evi["client port"] = packet.sport;
			evi["server port"] = packet.dport;
			PublishEvent("processed:tcp:new", evi, packet.parent);
			
			data.serveraddr = packet.dst;
			data.clientaddr = packet.src;
			data.serverport = packet.dport;
			data.clientport = packet.sport;
			data.inetname = packet.netname;
			
			data.client_stream.initial_sequence = packet.sequence + 1;
			return init2;
		}
		
		static uint64_t cid;
		
		FsmNode<TcpPacket &, TcpConnectionData> *init2;
};

uint64_t TcpConnectionFsmNodeInit1::cid = 0;

class TcpConnectionFsmNodeInit2 : public TcpConnectionFsmNode
{
	public:
		TcpConnectionFsmNodeInit2(FsmNode<TcpPacket &, TcpConnectionData> *pinit3) :
			init3(pinit3){}
		
	private:
		//Expect a SYN-ACK packet
		virtual FsmNode<TcpPacket &, TcpConnectionData> *ProcessCharacter(TcpPacket &packet, TcpConnectionData &data)
		{
			if (packet.fin || !packet.syn || packet.rst || packet.psh || !packet.ack || packet.urg){
				PublishEvent("anomaly:received:tcp:second packet not syn-ack", packet.ev, packet.parent);
				return 0;
			}
			
			if (packet.clientorigin){
				PublishEvent("anomaly:received:tcp:syn-ack from client", packet.ev, packet.parent);
				return 0;
			}
			
			if (packet.payload.GetLength()){
				PublishEvent("anomaly:received:tcp:second packet has payload", packet.ev, packet.parent);
				return 0;
			}
			
			data.client_stream.initial_sequence = packet.sequence + 1;
			return init3;
		}
		
		FsmNode<TcpPacket &, TcpConnectionData> *init3;
};

class TcpConnectionFsmNodeInit3 : public TcpConnectionFsmNode
{
	public:
		TcpConnectionFsmNodeInit3(FsmNode<TcpPacket &, TcpConnectionData> *pactive) :
			active(pactive){}
		
	private:
		//Expect an ACK packet
		virtual FsmNode<TcpPacket &, TcpConnectionData> *ProcessCharacter(TcpPacket &packet, TcpConnectionData &data)
		{
			if (!packet.clientorigin){
				PublishEvent("anomaly:received:tcp:initial ack from server", packet.ev, packet.parent);
				return 0;
			}
			
			if (packet.fin || packet.syn || packet.rst || packet.psh || !packet.ack || packet.urg){
				PublishEvent("anomaly:received:tcp:third packet not ack", packet.ev, packet.parent);
				return 0;
			}
			
			if (packet.payload.GetLength()){
				PublishEvent("anomaly:received:tcp:third packet has payload", packet.ev, packet.parent);
				return 0;
			}
			
			return active;
		}
		
		FsmNode<TcpPacket &, TcpConnectionData> *active;
};

class TcpConnectionFsmNodeActive : public TcpConnectionFsmNode
{
	public:
		TcpConnectionFsmNodeActive(FsmNode<TcpPacket &, TcpConnectionData> *pclient_fin,
			FsmNode<TcpPacket &, TcpConnectionData> *pserver_fin) :
			client_fin(pclient_fin), server_fin(pserver_fin){}
		
	private:
		//Expect ACK packets
		virtual FsmNode<TcpPacket &, TcpConnectionData> *ProcessCharacter(TcpPacket &packet, TcpConnectionData &data)
		{
			if (packet.syn){
				PublishEvent("anomaly:received:tcp:unexpected syn packet", packet.ev, packet.parent);
			}
			
			AddToBuffer(packet.payload, packet.clientorigin ? data.client_stream : data.server_stream, packet, data);
			
			if (packet.fin){
				return packet.clientorigin ? client_fin : server_fin;
			}
			
			return this;
		}
		
		FsmNode<TcpPacket &, TcpConnectionData> *client_fin;
		FsmNode<TcpPacket &, TcpConnectionData> *server_fin;
};

class TcpConnectionFsmNodeFinClient : public TcpConnectionFsmNode
{
	private:
		virtual FsmNode<TcpPacket &, TcpConnectionData> *ProcessCharacter(TcpPacket &packet, TcpConnectionData &data)
		{
			if (packet.syn){
				PublishEvent("anomaly:received:tcp:unexpected syn packet", packet.ev, packet.parent);
			}
			
			if (packet.clientorigin && packet.payload.GetLength()){
				PublishEvent("anomaly:received:tcp:unexpected payload after server fin", packet.ev, packet.parent);
			}
			
			AddToBuffer(packet.payload, packet.clientorigin ? data.client_stream : data.server_stream, packet, data);
			
			if (!packet.clientorigin && packet.fin){
				DynamicValue evi;
				evi["id"] = data.id;
				evi["server address"] = data.serveraddr;
				evi["client address"] = data.clientaddr;
				evi["server port"] = data.serverport;
				evi["client port"] = data.clientport;
				PublishEvent("processed:tcp:closed", evi, packet.parent);
				return 0; //we're done
			}
			
			return this; //TODO: Check if this is right?
		}
		
		
};

class TcpConnectionFsmNodeFinServer : public TcpConnectionFsmNode
{
	private:
		virtual FsmNode<TcpPacket &, TcpConnectionData> *ProcessCharacter(TcpPacket &packet, TcpConnectionData &data)
		{
			if (packet.syn){
				PublishEvent("anomaly:received:tcp:unexpected syn packet", packet.ev, packet.parent);
			}
			
			if (!packet.clientorigin && packet.payload.GetLength()){
				PublishEvent("anomaly:received:tcp:unexpected payload after server fin", packet.ev, packet.parent);
			}
			
			AddToBuffer(packet.payload, packet.clientorigin ? data.client_stream : data.server_stream, packet, data);
			
			if (packet.clientorigin && packet.fin){
				DynamicValue evi;
				evi["id"] = data.id;
				evi["server address"] = data.serveraddr;
				evi["client address"] = data.clientaddr;
				evi["server port"] = data.serverport;
				evi["client port"] = data.clientport;
				PublishEvent("processed:tcp:closed", evi, packet.parent);
				return 0; //we're done
			}
			
			return this; //TODO: Check if this is right?
		}
		
		
};

class TcpConnection
{
	public:
		TcpConnection() :
			active(&fin_client, &fin_server),
			init3(&active),
			init2(&init3),
			init1(&init2),
			fsm(TcpConnectionData(), &init1)
		{}
		
		bool ProcessPacket(TcpModule &parent, DynamicValue src, DynamicValue dst, std::string netname,
			Buffer payload, DynamicValue ev, bool clientorigin,
			uint16_t sport, uint16_t dport, uint32_t sequence, uint32_t acknowledgement, uint16_t window, uint16_t urgent,
			bool fin, bool syn, bool rst, bool psh, bool ack, bool urg)
		{
			TcpPacket packet(parent, src, dst, netname,
				payload, ev, clientorigin,
				sport, dport, sequence, acknowledgement, window, urgent,
				fin, syn, rst, psh, ack, urg);
			fsm.ProcessCharacter(packet);
			
			//return true if the current node is null
			//return false if the current node is not null
			//(the connection is deleted if we return true)
			return !fsm.GetNode();
		}
		
	private:
		TcpConnection(TcpConnection &);
		TcpConnection &operator=(TcpConnection &);
		
		TcpConnectionFsmNodeFinClient fin_client;
		TcpConnectionFsmNodeFinServer fin_server;
		TcpConnectionFsmNodeActive active;
		TcpConnectionFsmNodeInit3 init3;
		TcpConnectionFsmNodeInit2 init2;
		TcpConnectionFsmNodeInit1 init1;
		Fsm<TcpPacket &, TcpConnectionData> fsm;
};

class TcpModule : public Module
{
	public:
		TcpModule(Modules *modules, EventManager *manager, Configuration *configuration, Database *database) :
			Module("TCP", modules, manager, configuration, database)
		{
			std::list<std::string> ipv4tcp;
			ipv4tcp.push_back("processed:ipv4");
			ipv4tcp.push_back("received:tcp");
			GetManager()->Subscribe(GetNetworkStackEventSubscription<TcpModule>
				(ipv4tcp, this, &TcpModule::OnReceiveIpv4));
		}
		
		void PublishEvent(std::string name, DynamicValue ev)
		{
			Event e;
			e[name] = ev;
			GetManager()->Publish(e);
		}
		
	private:
		/*
		  http://www.ietf.org/rfc/rfc793.txt:
			The checksum also covers a 96 bit pseudo header conceptually
			prefixed to the TCP header.  This pseudo header contains the Source
			Address, the Destination Address, the Protocol, and TCP length.
		  
		  We'll need to implement this with a layer violation for IPv4 (and presumably
		  IPv6 too) - unfortunately.
		*/
		void OnReceiveIpv4(std::vector<Event> &events)
		{
			OnReceive("ipv4", events[0]["processed:ipv4"]["source"], events[0]["processed:ipv4"]["destination"],
				events[1]["received:tcp"]["data"]);
		}
		
		void OnReceive(std::string netname, DynamicValue src, DynamicValue dst, Buffer b)
		{
			Buffer bsrc = src;
			Buffer bdst = dst;
			
			DynamicValue ev;
			ev["network name"] = netname;
			ev["source address"] = src;
			ev["destination address"] = dst;
			ev["data"] = b;
			
			Buffer bnetname((const unsigned char *)netname.c_str(), netname.size());
			
			if (b.GetLength() < 20){
				PublishEvent("anomaly:received:tcp:packet length < header length", ev);
				return;
			}
			
			uint16_t sport = b.GetUInt16N(0);
			uint16_t dport = b.GetUInt16N(2);
			uint32_t sequence = b.GetUInt32N(4);
			uint32_t acknowledgement = b.GetUInt32N(8);
			uint8_t dataoffs = (b.GetUInt8(12) & 0xF0) >> 2;
			uint8_t flags1 = b.GetUInt8(13);
			uint8_t flags2 = b.GetUInt8(12);
			uint16_t window = b.GetUInt16N(14);
			uint16_t checksum = b.GetUInt16N(16);
			uint16_t urgent = b.GetUInt16N(18);
			
			bool fin = flags1 & 0x01;
			bool syn = flags1 & 0x02;
			bool rst = flags1 & 0x04;
			bool psh = flags1 & 0x08;
			bool ack = flags1 & 0x10;
			bool urg = flags1 & 0x20;
			bool ece = flags1 & 0x40;
			bool cwr = flags1 & 0x60;
			bool ns  = flags2 & 0x01;
			
			ev["source port"] = sport;
			ev["destination port"] = dport;
			ev["sequence"] = sequence;
			ev["acknowledgement"] = acknowledgement;
			ev["data offset"] = dataoffs;
			ev["window"] = window;
			ev["checksum"] = checksum;
			ev["urgent"] = urgent;
			
			ev["flags"]["fin"] = fin;
			ev["flags"]["syn"] = syn;
			ev["flags"]["rst"] = rst;
			ev["flags"]["psh"] = psh;
			ev["flags"]["ack"] = ack;
			ev["flags"]["urg"] = urg;
			ev["flags"]["ece"] = ece;
			ev["flags"]["cwr"] = cwr;
			ev["flags"]["ns"] = ns;
			
			bool processable = true;
			if (b.GetLength() < dataoffs){
				PublishEvent("anomaly:received:tcp:packet length < header length", ev);
				processable = false;
			}
			
			std::vector<DynamicValue> options;
			for (unsigned int i = 20; i < dataoffs;){
				DynamicValue option;
				uint8_t kind = b.GetUInt8(i);
				option["kind"] = kind;
				
				if (kind <= 1){
					i++;
				}
				else {
					if (dataoffs < i + 2){
						PublishOptionEvent("anomaly:received:tcp:incomplete option", ev, options, option, processable);
						break;
					}
					
					uint8_t length = b.GetUInt8(i + 1);
					option["length"] = length;
					
					if (length < 2){
						PublishOptionEvent("anomaly:received:tcp:invalid option length", ev, options, option, processable);
						break;
					}
					
					if (dataoffs < i + length){
						PublishOptionEvent("anomaly:received:tcp:incomplete option data", ev, options, option, processable);
						break;
					}
					
					option["data"] = Buffer(b, i + 2, length - 2);
					std::vector<DynamicValue> ol; //used in case 5
					switch (kind)
					{
						case 0:
							option["name"] = "end";
							break;
						case 1:
							option["name"] = "nop";
							break;
						case 2:
							option["name"] = "mss";
							if (length != 4){
								PublishOptionEvent("anomaly:received:tcp:invalid option:mss", ev, options, option, processable);
								break;
							}
							option["mss"] = b.GetUInt16N(i + 2);
							break;
						case 3:
							option["name"] = "ws";
							if (length != 3){
								PublishOptionEvent("anomaly:received:tcp:invalid option:ws", ev, options, option, processable);
								break;
							}
							option["ws"] = b.GetUInt8(i + 2);
							break;
						case 4:
							option["name"] = "sack permitted";
							if (length != 2){
								PublishOptionEvent("anomaly:received:tcp:invalid option:sack permitted", ev, options, option, processable);
								break;
							}
							break;
						case 5:
							option["name"] = "sack";
							if ((length != 10) && (length != 18) && (length != 26) && (length != 34)){
								PublishOptionEvent("anomaly:received:tcp:invalid option:sack", ev, options, option, processable);
								break;
							}
							
							for (unsigned int j = 2; j < length; j += 8){
								DynamicValue item;
								item["begin"] = b.GetUInt32N(i + j);
								item["end"] = b.GetUInt32N(i + j + 4);
								ol.push_back(item);
							}
							option["sacks"] = ol;
							
							break;
						case 8:
							option["name"] = "timestamp";
							if (length != 10){
								PublishOptionEvent("anomaly:received:tcp:invalid option:timestamp", ev, options, option, processable);
								break;
							}
							option["timestamp"] = b.GetUInt32N(i + 2);
							option["echo"] = b.GetUInt32N(i + 6);
							break;
						case 14:
							option["name"] = "checksum request";
							if (length != 3){
								PublishOptionEvent("anomaly:received:tcp:invalid option:checksum request", ev, options, option, processable);
								break;
							}
							option["checksum"] = b.GetUInt32N(i + 2);
							break;
						case 15:
							option["name"] = "checksum";
							option["checksum"] = Buffer(b, i + 2, length - 2);
							break;
						default:
							option["name"] = "unknown";
					}
					
					if (!processable){
						break;
					}
					
					i += length;
				}
				
				options.push_back(option);
			}
			
			if (!processable){
				return;
			}
			
			ev["options"] = options;
			
			Buffer payload = Buffer(b, dataoffs);
			ev["payload"] = payload;
			
			Buffer pseudoheader;
			if (netname == "ipv4"){
				uint8_t zero = 0;
				uint8_t protocol = 6;
				uint16_t length = dataoffs + 12;
				pseudoheader = bsrc + bdst +
					Buffer((const unsigned char *)&zero, 1) +
					Buffer((const unsigned char *)&protocol, 1) +
					Buffer((const unsigned char *)&length, 2);
				pseudoheader.SetUInt16N(10, length); //correct for network ordering
			}
			else {
				throw FormatError("OnReceive()'s pseudo header generator needs fixing");
			}
			//TODO: Check the checksum
			
			
			PublishEvent("processed:tcp:packet", ev); //A processed TCP packet with all the details intact
			
			bool stream = GetConfiguration()["tcp"]["stream"];
			
			if (stream){
				Buffer sdkey = bsrc + Buffer((const unsigned char *)&sport, 2) +
					bdst + Buffer((const unsigned char *)&dport, 2) + bnetname;
				Buffer dskey = bdst + Buffer((const unsigned char *)&dport, 2) +
					bsrc + Buffer((const unsigned char *)&sport, 2) + bnetname;
				
				if (connections.count(dskey)){
					ev["direction"] = std::string("server -> client");
					if (connections[dskey]->ProcessPacket(*this, src, dst, netname, payload, ev, false,
						sport, dport, sequence, acknowledgement, window, urgent, fin, syn, rst, psh, ack, urg)){
						delete connections[dskey];
						connections.erase(dskey);
					}
				}
				else {
					if (!connections.count(sdkey)){
						connections[sdkey] = new TcpConnection;
					}
					ev["direction"] = std::string("client -> server");
					if(connections[sdkey]->ProcessPacket(*this, src, dst, netname, payload, ev, true,
						sport, dport, sequence, acknowledgement, window, urgent, fin, syn, rst, psh, ack, urg)){
						delete connections[sdkey];
						connections.erase(sdkey);
					}
				}
			}
		}
		
		void PublishOptionEvent(std::string name, DynamicValue &ev, std::vector<DynamicValue> &options, DynamicValue &option, bool &processable)
		{
			options.push_back(option);
			ev["options"] = options;
			PublishEvent(name, ev);
			processable = false;
		}
		
		std::map<Buffer, TcpConnection *> connections;
};

void TcpConnectionFsmNode::PublishEvent(std::string name, DynamicValue evi, TcpModule &module)
{
	module.PublishEvent(name, evi);
}


INIT_MODULE(TcpModule);
