/*
	Hyperion - A network anomaly discovery system
	Copyright (C) 2011 Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "config.h"
#include "error.hpp"

#ifndef TIME_HPP
#define TIME_HPP

#include <stdint.h>
#include <sys/time.h>
#include <sys/timeb.h>

//TODO: Add string conversion methods?
//Before filling these in, check to make sure that we remove the int64_t representation of time
//Also check behaviours we want
//Also consider how to compare to zero, or to less than zero, etc (maybe static (const) zero value/function?)
//At some point, inline the methods that would benefit from being inlines
//I've inlined these because I felt it was not necessary to have them uninlined
class Time
{
	public:
		inline Time()
		{
			v = 0;
		}
		
		//Construct from the time in seconds
		inline Time(double t)
		{
			v = t * 1000000000.0;
		}
		
		inline Time(timeb t)
		{
			v = (int64_t)t.time * (int64_t)1000000000 + (int64_t)t.millitm * (int64_t)1000000;
		}
		
		inline Time(timeval t)
		{
			v = (int64_t)t.tv_sec * (int64_t)1000000000 + (int64_t)t.tv_usec * (int64_t)1000;
		}
		
		inline Time(timespec t)
		{
			v = (int64_t)t.tv_sec * (int64_t)1000000000 + (int64_t)t.tv_nsec;
		}
		
		static const Time zero;
		inline static Time s(int64_t t)
		{
			Time time;
			time.v = t * (int64_t)1000000000;
			return time;
		}
		
		inline static Time ms(int64_t t)
		{
			Time time;
			time.v = t * (int64_t)1000000;
			return time;
		}
		
		inline static Time us(int64_t t)
		{
			Time time;
			time.v = t * (int64_t)1000;
			return time;
		}
		
		inline static Time ns(int64_t t)
		{
			Time time;
			time.v = t;
			return time;
		}
		
		inline static Time now()
		{
			struct timeb t;
			ftime(&t);
			Time time;
			time.v = (int64_t)t.time * (int64_t)1000000000 + (int64_t)t.millitm * (int64_t)1000000;
			return time;
		}
		
		inline int64_t GetS()
		{
			return v / 1000000000;
		}
		
		inline int64_t GetMs()
		{
			return v / 1000000;
		}
		
		inline int64_t GetUs()
		{
			return v / 1000;
		}
		
		inline int64_t GetNs()
		{
			return v;
		}
		
		//This returns the time in seconds
		inline operator double()
		{
			return (double)v / 1000000000.0;
		}
		
		inline operator timeb()
		{
			timeb t;
			t.time = v/(int64_t)1000000000;
			t.millitm = (v % (int64_t)1000000000) / (int64_t)1000000;
			return t;
		}
		
		inline operator timeval()
		{
			timeval t;
			t.tv_sec = v / (int64_t)1000000000;
			t.tv_usec = (v % (int64_t)1000000000) / (int64_t)1000;
			return t;
		}
		
		inline operator timespec()
		{
			timespec t;
			t.tv_sec = v / (int64_t)1000000000;
			t.tv_nsec = v % (int64_t)1000000000;
			return t;
		}
		
		inline bool operator==(const Time &rhs) const
		{
			return v == rhs.v;
		}
		
		inline bool operator!=(const Time &rhs) const
		{
			return v != rhs.v;
		}
		
		inline bool operator<(const Time &rhs) const
		{
			return v < rhs.v;
		}
		
		inline bool operator>(const Time &rhs) const
		{
			return v > rhs.v;
		}
		
		inline bool operator<=(const Time &rhs) const
		{
			return v <= rhs.v;
		}
		
		inline bool operator>=(const Time &rhs) const
		{
			return v >= rhs.v;
		}
		
		inline Time operator+(const Time &rhs) const
		{
			Time t;
			t.v = v + rhs.v;
			return t;
		}
		
		inline Time operator-(const Time &rhs) const
		{
			Time t;
			t.v = v - rhs.v;
			return t;
		}
		
		inline Time operator*(const int64_t &rhs) const
		{
			Time t;
			t.v = v * rhs;
			return t;
		}
		
		inline Time operator/(const int64_t &rhs) const
		{
			Time t;
			t.v = v / rhs;
			return t;
		}
		
		inline Time operator*(const double &rhs) const
		{
			Time t;
			t.v = (double)v * rhs;
			return t;
		}
		
		inline Time operator/(const double &rhs) const
		{
			Time t;
			t.v = (double)v / rhs;
			return t;
		}
		
		inline Time &operator+=(const Time &rhs)
		{
			v += rhs.v;
			return *this;
		}
		
		inline Time &operator-=(const Time &rhs)
		{
			v -= rhs.v;
			return *this;
		}
		
		inline Time &operator*=(const int64_t &rhs)
		{
			v *= rhs;
			return *this;
		}
		
		inline Time &operator/=(const int64_t &rhs)
		{
			v /= rhs;
			return *this;
		}
		
		inline Time &operator*=(const double &rhs)
		{
			v = (double)v * rhs;
			return *this;
		}
		
		inline Time &operator/=(const double &rhs)
		{
			v = (double)v / rhs;
			return *this;
		}
		
	private:
		int64_t v;
};

#endif
